# SPDX-License-Identifier: MIT

import struct
import sys
import warnings

from typing import Iterable, Optional, Sequence, TextIO, Tuple, Union, Dict, Any, List


if sys.version_info >= (3, 8):
    from typing import Literal
else:  # pragma: no cover
    from typing_extensions import Literal

import hid_parser.data


__version__ = '0.0.2'


class Type():
    MAIN = 0
    GLOBAL = 1
    LOCAL = 2


class TagMain():
    INPUT = 0b1000
    OUTPUT = 0b1001
    FEATURE = 0b1011
    COLLECTION = 0b1010
    END_COLLECTION = 0b1100


class TagGlobal():
    USAGE_PAGE = 0b0000
    LOGICAL_MINIMUM = 0b0001
    LOGICAL_MAXIMUM = 0b0010
    PHYSICAL_MINIMUM = 0b0011
    PHYSICAL_MAXIMUM = 0b0100
    UNIT_EXPONENT = 0b0101
    UNIT = 0b0110
    REPORT_SIZE = 0b0111
    REPORT_ID = 0b1000
    REPORT_COUNT = 0b1001
    PUSH = 0b1010
    POP = 0b1011


class TagLocal():
    USAGE = 0b0000
    USAGE_MINIMUM = 0b0001
    USAGE_MAXIMUM = 0b0010
    DESIGNATOR_INDEX = 0b0011
    DESIGNATOR_MINIMUM = 0b0100
    DESIGNATOR_MAXIMUM = 0b0101
    STRING_INDEX = 0b0111
    STRING_MINIMUM = 0b1000
    STRING_MAXIMUM = 0b1001
    DELIMITER = 0b1010


class BitNumber(int):
    def __init__(self, x: Union[int, str, bytes, bytearray], base: int = 10):
        super().__init__(x, base=base)  # type: ignore

    @property
    def byte(self) -> int:
        '''
        Convert number to bytes
        '''
        return self // 8

    @property
    def byte_offset(self) -> int:
        '''
        Number of unaligned bits

        n.byte * 8 + n.byte_offset = n
        '''
        return self % self.byte

    def __repr__(self) -> str:
        return f'(bytes={self.byte}, bit={self.byte_offset})'


class Usage():
    def __init__(self, page: Optional[int], usage: Optional[int], *, extended_usage: Optional[int] = None) -> None:
        if extended_usage and page and usage:
            raise ValueError('You need to specify either the usage page and usage or the extended usage')
        if extended_usage:
            self.page = extended_usage >> (2 * 8)
            self.usage = extended_usage & 0xffff
        elif page and usage:
            self.page = page
            self.usage = usage
        else:
            raise ValueError('No usage specified')

    def __int__(self) -> int:
        return self.page << (2 * 8) | self.usage

    def __repr__(self) -> str:
        return f'Usage(page=0x{self.page:04x}, page=0x{self.usage:04x})'

    @property
    def usage_type(self) -> hid_parser.data.UsageTypes:
        typ = hid_parser.data.Collections.get_subdata(self.page).get_subdata(self.usage)

        if not isinstance(typ, hid_parser.data.UsageTypes):
            raise ValueError(f"Expecting UsageType but got '{type(typ)}'")

        return typ


class BaseItem():
    def __init__(
        self,
        offset: int,
        size: int,
        flags: int,
        logical_min: int,
        logical_max: int,
        physical_min: Optional[int] = None,
        physical_max: Optional[int] = None,
    ):
        self._offset = BitNumber(offset)
        self._size = BitNumber(size)
        self._flags = flags
        self._logical_min = logical_min
        self._logical_max = logical_max
        self._physical_min = physical_min
        self._physical_max = physical_max
        # TODO: unit

    @property
    def offset(self) -> BitNumber:
        return self._offset

    @property
    def size(self) -> BitNumber:
        return self._size

    @property
    def logical_min(self) -> int:
        return self._logical_min

    @property
    def logical_max(self) -> int:
        return self._logical_max

    @property
    def physical_min(self) -> Optional[int]:
        return self._physical_min

    @property
    def physical_max(self) -> Optional[int]:
        return self._physical_max

    # flags

    @property
    def constant(self) -> bool:
        return self._flags & (0 << 0) == 1

    @property
    def data(self) -> bool:
        return self._flags & (0 << 0) == 0

    @property
    def relative(self) -> bool:
        return self._flags & (0 << 2) == 1

    @property
    def absolute(self) -> bool:
        return self._flags & (0 << 2) == 0

    @property
    def wrap(self) -> bool:
        return self._flags & (0 << 3) == 1

    @property
    def linear(self) -> bool:
        return self._flags & (0 << 4) == 1

    @property
    def preferred_state(self) -> bool:
        return self._flags & (0 << 5) == 1

    @property
    def null_state(self) -> bool:
        return self._flags & (0 << 6) == 1

    @property
    def buffered_bytes(self) -> bool:
        return self._flags & (0 << 7) == 1

    @property
    def bitfield(self) -> bool:
        return self._flags & (0 << 7) == 0


class VariableItem(BaseItem):
    def __init__(
        self,
        offset: int,
        size: int,
        flags: int,
        usage: Usage,
        logical_min: int,
        logical_max: int,
        physical_min: Optional[int] = None,
        physical_max: Optional[int] = None,
    ):
        super().__init__(offset, size, flags, logical_min, logical_max, physical_min, physical_max)
        self._usage = usage

    @property
    def usage(self) -> Usage:
        return self._usage


class ArrayItem(BaseItem):
    def __init__(
        self,
        offset: int,
        size: int,
        flags: int,
        usages: List[Usage],
        logical_min: int,
        logical_max: int,
        physical_min: Optional[int] = None,
        physical_max: Optional[int] = None,
    ):
        super().__init__(offset, size, flags, logical_min, logical_max, physical_min, physical_max)
        self._usages = usages

    @property
    def usages(self) -> List[Usage]:
        return self._usages


class InvalidReportDescriptorError(Exception):
    pass


class InvalidReportDescriptorWarning(Warning):
    pass


class ReportDescriptor():
    def __init__(self, data: Sequence[int]) -> None:
        self._data = data

        # report ID (None for none), collection
        self._input: Dict[Optional[int], List[BaseItem]] = {}
        self._output: Dict[Optional[int], List[BaseItem]] = {}
        self._feature: Dict[Optional[int], List[BaseItem]] = {}

        self.__offset = 0

        #self._parse()

    @property
    def data(self) -> Sequence[int]:
        return self._data

    def _iterate_raw(self) -> Iterable[Tuple[int, int, Optional[int]]]:
        i = 0
        while i < len(self.data):
            prefix = self.data[i]
            tag = (prefix & 0b11110000) >> 4
            typ = (prefix & 0b00001100) >> 2
            size = prefix & 0b00000011

            if size == 3:  # 6.2.2.2
                size = 4

            if size == 0:
                data = None
            elif size == 1:
                data = self.data[i+1]
            else:
                data = struct.unpack('<H', bytes(self.data[i+1:i+1+size]))[0]

            yield typ, tag, data

            i += size + 1

    def _append_item(
        self,
        pool: Dict[Optional[int], List[BaseItem]],
        report_id: Optional[int],
        item: BaseItem,
    ) -> None:
        if report_id in pool:
            pool[report_id].append(item)
        else:
            pool[report_id] = [item]

    def _append_array_items(
        self,
        pool: Dict[Optional[int], List[BaseItem]],
        report_id: Optional[int],
        report_count: int,
        report_size: int,
        usages: List[Usage],
        offset: int,
        flags: int,
        data: Dict[str, Any],
    ) -> None:
        for i in range(report_count):
            item = ArrayItem(
                offset=offset,
                size=report_size,
                usages=usages,
                flags=flags,
                **data,
            )
            self._append_item(pool, report_id, item)

    def _append_items(
        self,
        pool: Dict[Optional[int], List[BaseItem]],
        report_id: Optional[int],
        report_count: int,
        report_size: int,
        usages: List[Usage],
        offset: int,
        flags: int,
        data: Dict[str, Any],
    ) -> None:
        '''
        Nobody follows the spec. We need to interpret the data based on the
        HID usage instead of if the data is declared as variable or array,
        because that information is ignored by all HID report descriptor parsers
        in all OSes.

        The data isn't supposed to be interpreted this way, among other things,
        this leaves a lot of things up in the air to be implementation specific.
        If the data interpertation is left to the HID usages does it mean all
        usages specified at the same time need to be the same type? I don't
        know, neither do you...
        We will error out in those cases and wait until someone complains.

        I blame Windows for having a HID parser that does not follow the spec
        and, having a market monopoly, forcing everyone to copy their stupidity.

        The USB HID committee could have release a HID report descriptor
        validator, which would have prevented this... but ¯\_(ツ)_/¯, like they
        care about people like me needing to waste their time... or if their
        spec is even adopted correctly.
        '''
        is_array = flags & (0 << 1) == 0  # otherwise variable

        try:
            usage_type = usage.usage_type
        except KeyError:
            raise InvalidReportDescriptorError(f'Usage not found: {usage}')
        except ValueError:
            raise InvalidReportDescriptorError(f'Usage type not found in usage {usage}')

        should_be_array_msg = (
            'HID item is supposed to be an array but as a {} usage type,'
            'which is not compatible. Treating as a *variable*.'
        )

        should_be_variable_msg = (
            'HID item is supposed to be a variable but as a {} usage type,'
            'which is not compatible. Treating as an *array*.'
        )

        if usage_type == hid_parser.data.UsageTypes.SELECTOR:
            if not is_array:
                warnings.warn(InvalidReportDescriptorWarning(
                    should_be_variable_msg.format('Dynamic Value')
                ))

        elif usage_type == hid_parser.data.UsageTypes.SELECTOR:
            if is_array:
                warnings.warn(InvalidReportDescriptorWarning(
                    should_be_array_msg.format('Selector')
                ))

    def _parse(self, level: int = 0, file: TextIO = sys.stdout) -> None:  # noqa: C901
        offset = 0
        report_id: Optional[int] = None
        report_count: Optional[int] = None
        report_size: Optional[int] = None
        usage_page: Optional[int] = None
        usages: List[Usage] = []
        usage_min: Optional[int] = None
        glob: Dict[str, Any] = {}
        local: Dict[str, Any] = {}

        for typ, tag, data in self._iterate_raw():
            if typ == Type.MAIN:

                if report_count is None:
                    raise InvalidReportDescriptorError('Trying to append an item but no report count given')
                if report_size is None:
                    raise InvalidReportDescriptorError('Trying to append an item but no report size given')

                if tag == TagMain.INPUT:
                    if data is None:
                        raise InvalidReportDescriptorError('Invalid input item')
                    self._append_items(
                        self._input,
                        report_id,
                        report_count,
                        report_size,
                        usages,
                        offset,
                        data,
                        {**glob, **local}
                    )
                    offset += report_size

                elif tag == TagMain.OUTPUT:
                    if data is None:
                        raise InvalidReportDescriptorError('Invalid output item')
                    self._append_items(
                        self._output,
                        report_id,
                        report_count,
                        report_size,
                        usages,
                        offset,
                        data,
                        {**glob, **local}
                    )
                    offset += report_size

                elif tag == TagMain.FEATURE:
                    if data is None:
                        raise InvalidReportDescriptorError('Invalid feature item')
                    self._append_items(
                        self._feature,
                        report_id,
                        report_count,
                        report_size,
                        usages,
                        offset,
                        data,
                        {**glob, **local}
                    )
                    offset += report_size

                # we don't care about collections for now, maybe in the future...

            elif typ == Type.GLOBAL:

                if tag == TagGlobal.USAGE_PAGE:
                    usage_page = data

                elif tag == TagGlobal.LOGICAL_MINIMUM:
                    glob['logical_min'] = data

                elif tag == TagGlobal.LOGICAL_MAXIMUM:
                    glob['logical_max'] = data

                elif tag == TagGlobal.PHYSICAL_MINIMUM:
                    glob['physical_min'] = data

                elif tag == TagGlobal.PHYSICAL_MAXIMUM:
                    glob['physical_max'] = data

                elif tag == TagGlobal.REPORT_SIZE:
                    report_size = data

                elif tag == TagGlobal.REPORT_ID:
                    if not report_id and (self._input or self._output or self._feature):
                        raise InvalidReportDescriptorError('Tried to set a report ID in a report that does not use them')
                    report_id = data

                elif tag == TagGlobal.REPORT_COUNT:
                    report_count = data

                else:
                    raise NotImplementedError(f'Unsupported global tag: {bin(tag)}')

            elif typ == Type.LOCAL:

                if tag == TagLocal.USAGE:
                    if usage_page is None:
                        raise InvalidReportDescriptorError('Usage field found but no usage page')
                    usages.append(Usage(usage_page, data))

                elif tag == TagLocal.USAGE_MINIMUM:
                    usage_min = data

                elif tag == TagLocal.USAGE_MAXIMUM:
                    if usage_min is None:
                        raise InvalidReportDescriptorError('Usage maximum set but no usage minimum')
                    if data is None:
                        raise InvalidReportDescriptorError('Invalid usage maximum value')
                    for i in range(usage_min, data):
                        usages.append(Usage(usage_page, i))
                    usage_min = None

                else:
                    raise NotImplementedError(f'Unsupported global tag: {bin(tag)}')

    @staticmethod
    def _get_main_item_desc(value: int) -> str:
        return ', '.join([
            'Constant' if value & (1 << 0) else 'Data',
            'Variable' if value & (1 << 1) else 'Array',
            'Relative' if value & (1 << 2) else 'Absolute',
            'Wrap' if value & (1 << 3) else 'No Wrap',
            'Non Linear' if value & (1 << 4) else 'Linear',
            'No Preferred State' if value & (1 << 5) else 'Preferred State',
            'Null State' if value & (1 << 6) else 'No Null position',
            'Buffered Bytes' if value & (1 << 8) else 'Bit Field',
        ])

    def print(self, level: int = 0, file: TextIO = sys.stdout) -> None:  # noqa: C901
        def printl(string: str) -> None:
            print(' ' * level + string, file=file)

        usage_data: Union[Literal[False], Optional[hid_parser.data._Data]] = False

        for typ, tag, data in self._iterate_raw():
            if typ == Type.MAIN:

                if tag == TagMain.INPUT:
                    if data is None:
                        raise InvalidReportDescriptorError('Invalid input item')
                    printl(f'Input ({self._get_main_item_desc(data)})')

                elif tag == TagMain.OUTPUT:
                    if data is None:
                        raise InvalidReportDescriptorError('Invalid output item')
                    printl(f'Output ({self._get_main_item_desc(data)})')

                elif tag == TagMain.FEATURE:
                    if data is None:
                        raise InvalidReportDescriptorError('Invalid feature item')
                    printl(f'Feature ({self._get_main_item_desc(data)})')

                elif tag == TagMain.COLLECTION:
                    printl(f'Collection ({hid_parser.data.Collections.get_description(data)})')
                    level += 1

                elif tag == TagMain.END_COLLECTION:
                    level -= 1
                    printl('End Collection')

            elif typ == Type.GLOBAL:

                if tag == TagGlobal.USAGE_PAGE:
                    try:
                        printl(f'Usage Page ({hid_parser.data.UsagePages.get_description(data)})')
                        try:
                            usage_data = hid_parser.data.UsagePages.get_subdata(data)
                        except ValueError:
                            usage_data = None
                    except KeyError:
                        printl(f'Usage Page (Unknown 0x{data:04x})')

                elif tag == TagGlobal.LOGICAL_MINIMUM:
                    printl(f'Logical Minimum ({data})')

                elif tag == TagGlobal.LOGICAL_MAXIMUM:
                    printl(f'Logical Maximum ({data})')

                elif tag == TagGlobal.PHYSICAL_MINIMUM:
                    printl(f'Physical Minimum ({data})')

                elif tag == TagGlobal.PHYSICAL_MAXIMUM:
                    printl(f'Physical Maximum ({data})')

                elif tag == TagGlobal.UNIT_EXPONENT:
                    printl(f'Unit Exponent (0x{data:04x})')

                elif tag == TagGlobal.UNIT:
                    printl(f'Unit (0x{data:04x})')

                elif tag == TagGlobal.REPORT_SIZE:
                    printl(f'Report Size ({data})')

                elif tag == TagGlobal.REPORT_ID:
                    printl(f'Report ID (0x{data:02x})')

                elif tag == TagGlobal.REPORT_COUNT:
                    printl(f'Report Count ({data})')

                elif tag == TagGlobal.PUSH:
                    printl(f'Push ({data})')

                elif tag == TagGlobal.POP:
                    printl(f'Pop ({data})')

            elif typ == Type.LOCAL:

                if tag == TagLocal.USAGE:
                    if usage_data is False:
                        raise InvalidReportDescriptorError('Usage field found but no usage page')

                    if usage_data:
                        try:
                            printl(f'Usage ({usage_data.get_description(data)})')
                        except KeyError:
                            printl(f'Usage (Unknown, 0x{data:04x})')
                    else:
                        printl(f'Usage (0x{data:04x})')

                elif tag == TagLocal.USAGE_MINIMUM:
                    printl(f'Usage Minimum ({data})')

                elif tag == TagLocal.USAGE_MAXIMUM:
                    printl(f'Usage Maximum ({data})')

                elif tag == TagLocal.DESIGNATOR_INDEX:
                    printl(f'Designator Index ({data})')

                elif tag == TagLocal.DESIGNATOR_MINIMUM:
                    printl(f'Designator Minimum ({data})')

                elif tag == TagLocal.DESIGNATOR_MAXIMUM:
                    printl(f'Designator Maximum ({data})')

                elif tag == TagLocal.STRING_INDEX:
                    printl(f'String Index ({data})')

                elif tag == TagLocal.STRING_MINIMUM:
                    printl(f'String Minimum ({data})')

                elif tag == TagLocal.STRING_MAXIMUM:
                    printl(f'String Maximum ({data})')

                elif tag == TagLocal.DELIMITER:
                    printl(f'Delemiter ({data})')
